use std::{collections::HashMap, sync::Arc};

use diesel::prelude::*;
use once_cell::sync::Lazy;
use parking_lot::{Mutex, RwLock, RwLockReadGuard, RwLockWriteGuard};
use rustracing_jaeger::Span;
use stretto::Cache;

use crate::{db::Inode, metrics};

#[derive(Clone)]
pub struct TracingLock(Arc<RwLock<()>>);
impl TracingLock {
    pub fn read(&self, span: &Span) -> RwLockReadGuard<'_, ()> {
        let _span = metrics::span("get_read_lock", span);
        self.0.read()
    }

    pub fn write(&self, span: &Span) -> RwLockWriteGuard<'_, ()> {
        let _span = metrics::span("get_write_lock", span);
        self.0.write()
    }
}

pub static DB_MANAGER: Lazy<DBManager> = Lazy::new(DBManager::new);

pub struct DBManager {
    locks: Mutex<HashMap<i32, Arc<RwLock<()>>>>,
    node_cache: Cache<i32, Inode>,
    children_cache: Cache<i32, Vec<Inode>>
}

impl DBManager {
    fn new() -> Self {
        Self {
            locks: Mutex::new(HashMap::new()),
            node_cache: Cache::new(10000, 1000).expect("Failed to create node cache"),
            children_cache: Cache::new(1000, 100).expect("Failed to create child cache")
        }
    }

    pub fn create_node(
        &self,
        span: &Span,
        db: &mut super::RawDBConnection,
        file: bool,
        _name: String,
        parent: Option<i32>,
        owner: i32
    ) -> Inode {
        use crate::schema::inode::dsl::*;
        let _span = metrics::span("insert_node_into_db", span);
        let node: Inode = diesel::insert_into(inode)
            .values(crate::db::inode::NewInode {
                is_file: file,
                name: _name,
                parent_id: parent,
                owner_id: owner,
                size: None,
                has_preview: false
            })
            .get_result(db)
            .expect("Failed to insert new inode");

        self.node_cache.insert(node.id, node.clone(), 1);
        if let Some(parent) = parent {
            self.children_cache.remove(&parent);
        }

        let owner = node.owner_id.to_string();
        if file {
            metrics::NODES.with_label_values(&["file", owner.as_str()]).inc();
        } else {
            metrics::NODES.with_label_values(&["folder", owner.as_str()]).inc();
        }

        node
    }

    pub fn get_node(&self, span: &Span, db: &mut super::RawDBConnection, node_id: i32) -> Option<Inode> {
        use crate::schema::inode::dsl::*;
        let inner_span = metrics::span("get_node", span);
        let node = self.node_cache.get(&node_id);
        match node {
            Some(v) => {
                metrics::CACHE.node.hit.inc();
                Some(v.value().clone())
            }
            None => {
                let _span = metrics::span("get_node_from_db", &inner_span);
                metrics::CACHE.node.miss.inc();
                let v: Inode = inode.find(node_id).first(db).ok()?;
                self.node_cache.insert(node_id, v.clone(), 1);
                Some(v)
            }
        }
    }

    pub fn get_children(&self, span: &Span, db: &mut super::RawDBConnection, node_id: i32) -> Vec<Inode> {
        use crate::schema::inode::dsl::*;
        let inner_span = metrics::span("get_children", span);
        let children = self.children_cache.get(&node_id);
        match children {
            Some(v) => {
                metrics::CACHE.children.hit.inc();
                v.value().clone()
            }
            None => {
                let _span = metrics::span("get_children_from_db", &inner_span);
                metrics::CACHE.children.miss.inc();
                let v = inode.filter(parent_id.eq(node_id)).load(db).expect("Failed to get children of node");
                self.children_cache.insert(node_id, v.clone(), 1);
                v
            }
        }
    }

    pub fn save_node(&self, span: &Span, db: &mut super::RawDBConnection, node: &Inode) {
        let _span = metrics::span("save_node_to_db", span);
        self.node_cache.insert(node.id, node.clone(), 1);
        if let Some(p) = node.parent_id {
            self.children_cache.remove(&p);
        }
        diesel::update(node).set(node.clone()).execute(db).expect("Failed to save node");
    }

    pub fn delete_node(&self, span: &Span, db: &mut super::RawDBConnection, node: &Inode) {
        let inner_span = metrics::span("delete_node", span);
        let owner = node.owner_id.to_string();
        if node.is_file {
            let _span = metrics::span("delete_node_files", &inner_span);
            let file_name = format!("./files/{}", node.id);
            let file = std::path::Path::new(&file_name);
            let preview_name = format!("./files/{}_preview.jpg", node.id);
            let preview = std::path::Path::new(&preview_name);
            if file.exists() {
                std::fs::remove_file(file).expect("Failed to delete file");
            }
            if preview.exists() {
                std::fs::remove_file(preview).expect("Failed to delete preview");
            }
            metrics::NODES.with_label_values(&["file", owner.as_str()]).dec();
            metrics::DISK_USAGE.with_label_values(&[owner.as_str()]).sub(node.size.unwrap_or(0));
        } else {
            metrics::NODES.with_label_values(&["folder", owner.as_str()]).dec();
        }

        let _span = metrics::span("delete_node_from_db", &inner_span);
        diesel::delete(node).execute(db).expect("Failed to delete node");
        self.node_cache.remove(&node.id);
        self.children_cache.remove(&node.id);
        if let Some(p) = node.parent_id {
            self.children_cache.remove(&p);
        }
    }

    pub fn get_lock(&self, user: i32) -> TracingLock {
        TracingLock(self.locks.lock().entry(user).or_insert_with(|| Arc::new(RwLock::new(()))).clone())
    }
}
