pub mod basic;
pub mod gitlab;
pub mod tfa;

use std::ops::Add;

use once_cell::sync::Lazy;
use ring::{rand, rand::SecureRandom};
use rustracing_jaeger::Span;

#[derive(Debug, serde::Deserialize, serde::Serialize)]
pub struct JWTClaims {
    pub exp: i64,
    pub iat: i64,
    pub jti: i32,
    pub sub: i32
}

pub static JWT_ALGORITHM: jsonwebtoken::Algorithm = jsonwebtoken::Algorithm::HS512;
pub static SEC_RANDOM: Lazy<rand::SystemRandom> = Lazy::new(rand::SystemRandom::new);
pub static JWT_SECRET: Lazy<Vec<u8>> = Lazy::new(get_jwt_secret);
pub static JWT_DECODE_KEY: Lazy<jsonwebtoken::DecodingKey> =
    Lazy::new(|| jsonwebtoken::DecodingKey::from_secret(JWT_SECRET.as_slice()));
pub static JWT_ENCODE_KEY: Lazy<jsonwebtoken::EncodingKey> =
    Lazy::new(|| jsonwebtoken::EncodingKey::from_secret(JWT_SECRET.as_slice()));

fn get_jwt_secret() -> Vec<u8> {
    let secret = std::fs::read("jwt.secret");
    if let Ok(secret) = secret {
        secret
    } else {
        let mut secret: [u8; 128] = [0; 128];
        SEC_RANDOM.fill(&mut secret).expect("Failed to generate jwt secret");
        std::fs::write("jwt.secret", secret).expect("Failed to write jwt secret");
        Vec::from(secret)
    }
}

pub fn get_token(span: &Span, user: &crate::db::User, db: &mut crate::db::DBConnection) -> String {
    let iat = chrono::Utc::now();
    let exp = iat.add(chrono::Duration::hours(24)).timestamp();
    let iat = iat.timestamp();

    let token = db.create_token(span, user.id, exp);

    let claims = JWTClaims {
        exp,
        iat,
        jti: token.id,
        sub: user.id
    };

    jsonwebtoken::encode(
        &jsonwebtoken::Header::new(JWT_ALGORITHM),
        &claims,
        &JWT_ENCODE_KEY
    )
    .expect("Failed to create JWT token")
}

pub fn hash_password(password: &String) -> String {
    let mut salt = [0_u8; 16];
    SEC_RANDOM.fill(&mut salt).expect("Failed to generate salt");
    let config = argon2::Config {
        mem_cost: 64 * 1024,
        variant: argon2::Variant::Argon2id,
        ..Default::default()
    };
    argon2::hash_encoded(password.as_bytes(), &salt, &config).expect("Failed to hash password")
}
